/*
 *    / __/ /
 *   / _// / . ||//.\/  ./
 *  /_/ /_/_/|_|_/\_/_/\_\
 *  http://www.personal.triticom.com/~erm/GP32/
 *	FileSel.c by Flavor (1-28-04)
 *
 *    This is the basis for a fairly simple directory browser and file selector.
 *    You will need to edit some of the code for your specific application.
 *    To use this, set up all the stuff that you want to change, and then call
 *        runFileSel()
 *    When the user has selected a file, then processFilename() will be called.
 *
 *    WARNING:  runFileSel() calls GpFatInit() so if you call it in your code,
 *              you may want to remove this additional call.
 */

#include "gpmain.h"
#include "FileSel.h"
#include "z80.h"

//if you want to change it so that the user can only browse a certain sub directory,
//  then set this up (but you may want to add some testing to see if it exists first)
#define ROOT_DIRECTORY "gp:\\gpmm\\gp81\\"

//NUM_SUPPORTED_EXTENSIONS has to be the number of supported extensions
// if you put this to 0, then only directories will be displayed
// if you set this to -1, then all files will be supported
#define NUM_SUPPORTED_EXTENSIONS 1
#if NUM_SUPPORTED_EXTENSIONS > 0
 //these must be all lower case!!!
 char supportedExtensions[NUM_SUPPORTED_EXTENSIONS][4] =
 {
	"ppp"
 };
#endif


//if you have errors about the Quit function not being defined, try uncommenting this next line
#define Quit(m) while(1);

//these are all set up just before processFilename is called
char fileNames[MAX_NUM_FIES][FILENAME_LEN];
int numFilenames = 0; 
int currFilename = 0;

//this processes the selected file (when a user finally selects a file, then this is called)
//  when this gets called, fileNames, numFilenames, and currFilename will all bet set up
//  also, the relative path will be set so that you will only need to open the file to use it
//  I suggest calling a function that you have elsewhere so you don't fill this function up with
//  your application specific code
void processFilename()
{
  strncat(loadgame,fileNames[currFilename],strlen(fileNames[currFilename])-4); // remove extension
}


//displayDirectory displays the current directory and its supported files
//  If you want to make the file list view different, then you will need to change this function
//
//  If you use my code, then BYTES_PER_PIXEL must be set to either 1 or 2 
//  depending on if you are set to use 8 or 16 bit graphics.
//  graphmem is a pointer to the back buffer
//  drawMessage draws text at a certain X,Y (to the back buffer)
//  Flip displays the back buffer on the screen
//  
void displayDirectory(fileStruct *files, int numFiles, int currSelection, char *path)
{
	int x, y;
	unsigned char r, g, b;
	int color;
	char message[50];
	int i;

	//      DrawMainScreen(thecategory,0);
      GpRectFill(NULL, &vptr[0], 20, 30, 285, 175, 0);
      GpRectFill(NULL, &vptr[0], 21, 31, 283, 173, 255);

      GpRectFill(NULL, &vptr[0], 30, 21, 145, 18, 0);
      GpRectFill(NULL, &vptr[0], 31, 22, 143, 16, 255);

      GpTextOut(NULL, &vptr[0], 33,23,"Select zx81 file",0);

	//displey the path

	sprintf(message, "Viewing %s", path);
      	GpTextOut(NULL,&vptr[0],33, 40, message, 0);

	//display the files/directories that are above/below the currSelection 
	for(i=currSelection-3; i <= currSelection+3; i++)
	{
		if(i >= 0 && i < numFiles && i != currSelection)
		{
			if(files[i].attrib == ATTR_DIR)  //if it's a DIR, we put < and > around it
			{
				sprintf(message, "<%s>", files[i].filename);
				GpTextOut(NULL,&vptr[0],(320-(14*8))/2, 240/2+(20*(i-currSelection)), message, 0);
			}
			else  //otherwise, we just display it normal
				GpTextOut(NULL,&vptr[0],(320-(14*8))/2, 240/2+(20*(i-currSelection)), files[i].filename, 0);
		}
	}

	//display the actual currSelection
	if(files[currSelection].attrib == ATTR_DIR)  //if it's a DIR, we put < and > around it
	{
		sprintf(message, "<%s>", files[currSelection].filename);
		GpTextOut(NULL,&vptr[0],(320-(14*8))/2, 240/2, message, 224);
	}
	else   //otherwise, we just display it normal
		GpTextOut(NULL,&vptr[0],(320-(14*8))/2, 240/2, files[currSelection].filename, 224);
	
}



/*
 *
	***   ***   ***   ***   ***   ***   ***   ***   ***   ***   ***   ***
	You should not need to change any of the functions beyond this point!
	***   ***   ***   ***   ***   ***   ***   ***   ***   ***   ***   ***
 *
 */

//this tests to see if a file extension is supported
int supportedFileExt(char * filename)
{
#if NUM_SUPPORTED_EXTENSIONS == -1
	return TRUE;
#elif NUM_SUPPORTED_EXTENSIONS == 0
	return FALSE;
#else

	int len = strlen(filename);
	char tempEXT[FILENAME_LEN];
	int i, j;

	for(i=0;i<len;i++)
	{
		if(filename[i] == '.')
			break;
	}
	if(i==len)
		return FALSE;  //there wasn't even a period in the filename

	i++;//move to the first char after the period
	for(j=0;i<len;i++,j++)
		tempEXT[j] = tolower(filename[i]);
	tempEXT[j]=0;//null terminate

	for(i=0; i < NUM_SUPPORTED_EXTENSIONS; i++)
	{
			if(strcmp(supportedExtensions[i], tempEXT) == 0)
			return TRUE;
	}
	return FALSE;
#endif
}

//this gets the current state of the GP32's keys
word GetJoypadWord()
{
	int ExKey;
	GpKeyGetEx(&ExKey);
	return (ExKey & 0xFFFF);
}

//this processes the user's input
int userSelection(fileStruct *files, int numFiles, int currSelection, char *path)
{
	int needDisp = TRUE;
	word joypad, buttonsProcessed = 0, prevButtonsProcessed = 0;

	while(1)
	{
		joypad = GetJoypadWord();

		/*
		#define	GPC_VK_NONE		0x00
		#define	GPC_VK_LEFT		0x01
		#define	GPC_VK_UP		0x08	
		#define	GPC_VK_RIGHT		0x04
		#define	GPC_VK_DOWN		0x02

		#define GPC_VK_START		0x100
		#define GPC_VK_SELECT		0x200

		#define GPC_VK_FA		GPC_VK_ENTER
		#define GPC_VK_FB		GPC_VK_F3
		#define GPC_VK_FL		GPC_VK_F2
		#define GPC_VK_FR		GPC_VK_F1
		*/

		if(joypad & GPC_VK_DOWN && currSelection < numFiles-1)
		{
			buttonsProcessed |= GPC_VK_DOWN;

			if(currSelection < numFiles-1)
				currSelection++;
			else
				currSelection = 0;

			needDisp = TRUE;
		}
		if(joypad & GPC_VK_UP && currSelection > 0)
		{
			buttonsProcessed |= GPC_VK_UP;
			if(currSelection > 0)
				currSelection--;
			else
				currSelection = numFiles-1;

			needDisp = TRUE;
		}
		if(needDisp)
		{
			displayDirectory(files, numFiles, currSelection, path);
			needDisp = FALSE;
		}
		
		if(joypad & GPC_VK_FA)
		{
			buttonsProcessed |= GPC_VK_FA;
			while(GetJoypadWord() & buttonsProcessed);
			return currSelection;
		}
		
		if(joypad & GPC_VK_FB)
		{
			buttonsProcessed |= GPC_VK_FB;
			while(GetJoypadWord() & buttonsProcessed);
			return -1;
		}

		if(joypad & GPC_VK_SELECT)
		{
			buttonsProcessed |= GPC_VK_SELECT;
			while(GetJoypadWord() & buttonsProcessed);
			return -2;
		}


		
		if (GpSMCDetected ()==2) {
		  return -2;
		}
		
		unsigned int time0, time1;
		time0 = GpTickCountGet();
		while(GetJoypadWord() & buttonsProcessed)
		{
			time1 = GpTickCountGet() - time0;
			if(prevButtonsProcessed != buttonsProcessed && time1 > 400)
				break;
			if(prevButtonsProcessed == buttonsProcessed && time1 > 100)
				break;
		};
		prevButtonsProcessed = buttonsProcessed;
		buttonsProcessed = 0;
	}
}


//this compares two fileStruct entries to see which one should come first (sorting)
int fileStructCompare (const void * a, const void * b)
{
	fileStruct *fsA, *fsB;

	fsA = (fileStruct *) a;
	fsB = (fileStruct *) b;

	if (fsA->attrib == ATTR_DIR && fsB->attrib != ATTR_DIR)
		return -1;
    else if (fsA->attrib != ATTR_DIR && fsB->attrib == ATTR_DIR)
		return 1;
	else
		return strcmp(fsA->filename, fsB->filename);
}

//takes an array of GPDIRENTRYs and turns them into an array of fileStructs
int plistToFiles(GPDIRENTRY *p_list, int read_count, fileStruct *files)
{
	int i;
	ERR_CODE err_code;
	int numFiles = 0;
	GPFILEATTR mAttr;

	for(i=0;i<read_count;i++)  //add all the supported files to the list
	{
		if(strcmp(p_list[i].name, "..") == 0 || strcmp(p_list[i].name, ".") == 0)
			continue;

		err_code = GpFileAttr(p_list[i].name, &mAttr);
		if (err_code != SM_OK)
		{
			Quit("GpFileAttr ERROR");
		}

		if(mAttr.attr & 0x10)
		{
			strncpy(files[numFiles].filename, p_list[i].name, 13);
			files[numFiles].attrib = ATTR_DIR;
			numFiles++;
		}
		else
		{
			if(supportedFileExt(p_list[i].name))
			{
				strncpy(files[numFiles].filename, p_list[i].name, 13);
				files[numFiles].attrib = ATTR_FILE;
				numFiles++;
			}
		}		
	}

	//now sort the files
	qsort (files, numFiles, sizeof(fileStruct), fileStructCompare);

	return numFiles;
}

//the main directory getting function  (recursive)
int getDir(char *basePath, char *dirChoice)
{
	char localPath[PATH_LEN];
	fileStruct files[MAX_NUM_FIES];
	int numFiles = 0;
	int selection = 0;
	int prevSelection = 0;
	ERR_CODE err_code;
	GPDIRENTRY *p_list;
	int atRoot = FALSE;
	unsigned long list_count = 100, read_count = 0;
	int i;

	if(dirChoice == NULL)
	{
		atRoot = TRUE;
		strncpy(localPath, basePath, PATH_LEN);
	}
	else
	{
		strncpy(localPath, basePath, PATH_LEN);
		strcat(localPath, dirChoice);
		strcat(localPath, "\\");
	}


	GpRelativePathSet(localPath);
	
	err_code = GpDirEnumNum(localPath, &list_count);
	if (err_code != SM_OK)
	{
		//sprintf(message, "GpDirEnumNum ERR=%d %s", err_code, localPath);
		//Quit(message);
		Quit("GpDirEnumNum ERROR");
	}
	
	p_list = (GPDIRENTRY*)gp_mem_func.calloc(list_count, sizeof(GPDIRENTRY));

	GpDirEnumList(localPath, 0, list_count, p_list, &read_count);

	numFiles = plistToFiles(p_list, read_count, files);

	gp_mem_func.free(p_list);	


	if(numFiles == 0)
		return FALSE;

	while(1)
	{
		prevSelection = selection;
		selection = userSelection(files, numFiles, selection, localPath);

		if(selection == -2)
		  {
		    return -2;
		  }
		if(selection == -1 && !atRoot)
		{
			return FALSE;
		}
		else if(files[selection].attrib == ATTR_DIR)
		{
		  int gotdir;
		  gotdir=getDir(localPath, files[selection].filename);
		  if (gotdir)
				return TRUE;
		  if (gotdir==-2) { return -2; }
			//otherwise, restore our local
			GpRelativePathSet(localPath);
		}
		else if(files[selection].attrib == ATTR_FILE) //it's a file
		{
			//add all the supported files in the current directory to a global list
			numFilenames = currFilename = 0;
			for(i=0;i<numFiles;i++)
			{
				if(files[i].attrib == ATTR_FILE)
				{
					if(i == selection)
						currFilename = numFilenames;
					strcpy(fileNames[numFilenames], files[i].filename);
					numFilenames++;
				}
			}
			
			processFilename();
			return -2;
			//reset selection
			for(i=0;i<numFiles;i++)
			{
				if(strcmp(files[i].filename, fileNames[currFilename]) == 0)
				{
					selection = i;
				}
			}
		}
		else  //try again
		{
			selection = prevSelection;
		}
	}
}

//the entry point
void runFileSel()
{
  //	GpFatInit();  //you may want to comment this out if you already called it in your code
	GpRelativePathSet(ROOT_DIRECTORY);

	//for the first call to getDir, the second parameter must be NULL!
	//  the first parameter can be any directory, but then the user may not go above that directory
	//  effectively, the first parameter is the virtual "root" of the browser
	if(!getDir(ROOT_DIRECTORY, NULL))    
		Quit("No Files Found!");
}
