/* gp81, a port of z81 for the gp32.
 * Copyright (C) 1994 Ian Collier. Z81 changes (C) 1995-6 Russell Marks.
 * gp81 port by Woogal (woogal@sector808.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>

#include <rawkey.h>
#include "z80.h"
#include "gpmain.h"
#include "gp32.h"
#include "zx81kybd.h"
#include "FileSel.h"

int but_a,but_b,but_l,but_r,but_u,but_d,but_ls,but_rs;

unsigned long rGLOBALCOUNTER;
unsigned long lastinttime;

/* remember, this table is ignoring shifts... */
static struct {unsigned char port,mask;} keytable[]={
/*  SP      !        "        #        $        %        &        '  */
{7,0xfe},{8,0xff},{3,0xfd},{8,0xff},{5,0xf7},{8,0xff},{8,0xff},{8,0xff},
/*  (       )        *        +        ,        -        .        /  */
{5,0xfb},{5,0xfd},{4,0xfb},{6,0xfb},{7,0xfd},{6,0xf7},{7,0xfd},{0,0xef},
/*  0       1        2        3        4        5        6        7  */
{4,0xfe},{3,0xfe},{3,0xfd},{3,0xfb},{3,0xf7},{3,0xef},{4,0xef},{4,0xf7},
/*  8       9        :        ;        <        =        >        ?  */
{4,0xfb},{4,0xfd},{0,0xfd},{0,0xfb},{7,0xf7},{6,0xfb},{7,0xfb},{0,0xef},
/*  @       A        B        C        D        E        F        G  */
{8,0xff},{1,0xfe},{7,0xef},{0,0xf7},{1,0xfb},{2,0xfb},{1,0xf7},{1,0xef},
/*  H       I        J        K        L        M        N        O  */
{6,0xef},{5,0xfb},{6,0xf7},{6,0xfb},{6,0xfd},{7,0xfb},{7,0xf7},{5,0xfd},
/*  P       Q        R        S        T        U        V        W  */
{5,0xfe},{2,0xfe},{2,0xf7},{1,0xfd},{2,0xef},{5,0xf7},{0,0xef},{2,0xfd},
/*  X       Y        Z        [        \        ]        ^        _  */
{0,0xfb},{5,0xef},{0,0xfd},{8,0xff},{8,0xff},{8,0xff},{8,0xff},{8,0xff},
/*  `       a        b        c        d        e        f        g  */
{8,0xff},{1,0xfe},{7,0xef},{0,0xf7},{1,0xfb},{2,0xfb},{1,0xf7},{1,0xef},
/*  h       i        j        k        l        m        n        o  */
{6,0xef},{5,0xfb},{6,0xf7},{6,0xfb},{6,0xfd},{7,0xfb},{7,0xf7},{5,0xfd},
/*  p       q        r        s        t        u        v        w  */
{5,0xfe},{2,0xfe},{2,0xf7},{1,0xfd},{2,0xef},{5,0xf7},{0,0xef},{2,0xfd},
/*  x       y        z        {        |        }        ~       DEL */
{0,0xfb},{5,0xef},{0,0xfd},{8,0xff},{8,0xff},{8,0xff},{8,0xff},{8,0xff}
};

#define MAX_KEYHOLD 8

int keyhold=MAX_KEYHOLD;

void RTCInt(void) __attribute__ ((interrupt ("IRQ")));

unsigned char kports[9]={0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff};
char tmpstr[100];
  char keypressed;
int capson=0;
unsigned char *dfile;
unsigned char *memptr[4]={mem,mem+0x4000,mem+0x8000,mem+0xc000};
unsigned char *iptr=mem;

unsigned long tstates=0,tsmax=(1<<30);
int hires=0,hiscrn;
int help=0;
int topspeed=0,lowref=0;

int memattr[4]={0,2,1,1};

int screen_dirty;
int hsize=256,vsize=192;
volatile int interrupted=0;
int forceok=0;
int scrn_freq=2;
int booting=1;
int input_wait=0;

unsigned char chrmap_old[24*32],chrmap[24*32];
unsigned char himap_old[192*32],himap[192*32];

int refresh_screen=1;

void initmem() {
  loadrom(mem);
  hacks();
  gm_memset(mem+16384,0xff,49152);
  gm_memset(chrmap_old,0xff,768);
}

void GpError(char *text){
    gm_memset(vptr[0].ptbuffer, 0xFF, 320*240);
    GpTextOut(NULL, &vptr[0], 10, 10 , text, 0x00);
}

// RTC
// idea and first written, Jouni Korhonen
// rewritten and extended (c)2004 Mirko Roller

void RTCInt(void) {
  rGLOBALCOUNTER++;
  if ((int)((rGLOBALCOUNTER/64.0)*50)>lastinttime) { // change timing to 1/50 instead of 1/64
    lastinttime=(int)((rGLOBALCOUNTER/64.0)*50);
    if (interrupted<2) interrupted=1;
  }
}


void setupRTCInt( void ) {
  // enable RTC
  rCLKCON |= 0x800;
  //
  // The CPU clock indepentent RTC Timer
  //
  // The ticnt register 0x15700044 (32 bit) offers:
  //
  // BIT   7: 0=disable 1=enable timer
  // BIT 0-6: tick time count value 1-127
  //
  // %10000000 should result in a 1/128 timer, but is not working
  // %10000001 result in 1/64 timer
  // %11111111 result in 1/1  timer

  rTICINT = 0x81;
  //(*(volatile unsigned char *)rTICINT) = 0x81;
  rGLOBALCOUNTER = 0;
}


void InstallRTC( void ) {
  gp_disableIRQ();
  gp_installSWIIRQ(8,RTCInt);
  gp_enableIRQ();
}

unsigned long gp_getRTC() {
  return rGLOBALCOUNTER;
}

void gp_initRTC() {
  setupRTCInt();
  InstallRTC();
}

void gp_clearRTC() {
  rGLOBALCOUNTER=0;
}

void GpSetPaletteEntry ( int i,int r, int g, int b )
{
  GP_PALETTEENTRY entry = GP_RGB24(r,g,b);
  GpPaletteEntryChange ( i, 1, &entry, 0 );
}


int update_gpkbd() {
  int cbch;
  cbch=1;

    cbch = pollKbdKey();
    if (cbch=='b') { // right arrow (use as ctrl)
      kports[0]&=0xfe;
      keypressed=cbch;
      keyhold=MAX_KEYHOLD;
      return 1;
    }
    if (cbch=='n') { // newline
      kports[6]&=0xfe;
      keypressed=cbch;
      keyhold=MAX_KEYHOLD;
      return 1;
    }
    if (((cbch-=32)>=0) && (cbch<128)) { // normal key
      kports[keytable[cbch].port]&=keytable[cbch].mask;
      keypressed=cbch+32;
      keyhold=MAX_KEYHOLD;
    }
}

int keyrelease() {
    if (keypressed=='b') { // right arrow (use as ctrl)
      return 1;
    }
    if (keypressed=='n') { // newline
      kports[6]|=1;
      return 1;
    }
    if (((keypressed-=32)>=0) && (keypressed<128)) { // normal key
      kports[keytable[keypressed].port]|=~keytable[keypressed].mask;
      kports[0]|=1;
    }
  keypressed=0;
}

int is_key_pressed(char key)
{
    if (keypressed==key) {
      if (capson) capson=0;
      return 1;
    }
    return 0;
}

void sighandler(a)
int a;
{
if(interrupted<2) interrupted=1;
}


void dontpanic(a)
int a;
{
exit(1);
}

void update_gppress(int ExKey)
{
  int key;

  if (ExKey & GPC_VK_LEFT) {
      kports[keytable[but_l-32].port]&=keytable[but_l-32].mask;
      keypressed=but_l;
  }
  if (ExKey & GPC_VK_RIGHT) {
      kports[keytable[but_r-32].port]&=keytable[but_r-32].mask;
      keypressed=but_r;
  }
  if (ExKey & GPC_VK_DOWN) {
      kports[keytable[but_d-32].port]&=keytable[but_d-32].mask;
      keypressed=but_d;
  }
  if (ExKey & GPC_VK_UP) {
      kports[keytable[but_u-32].port]&=keytable[but_u-32].mask;
      keypressed=but_u;
  }
  if (ExKey & GPC_VK_FA) {
      kports[keytable[but_a-32].port]&=keytable[but_a-32].mask;
      keypressed=but_a;
  }
  if (ExKey & GPC_VK_FB) {
      kports[keytable[but_b-32].port]&=keytable[but_b-32].mask;
      keypressed=but_b;
  }
  if (ExKey & GPC_VK_FL) {
      kports[keytable[but_ls-32].port]&=keytable[but_ls-32].mask;
      keypressed=but_ls;
  }
  if (ExKey & GPC_VK_FR) {
      kports[keytable[but_rs-32].port]&=keytable[but_rs-32].mask;
      keypressed=but_rs;
  }

  if (ExKey & GPC_VK_START) {
    GpKeyGetEx(&key);while(key!=0)GpKeyGetEx(&key);
    if(help)
      memset(vptr[0].ptbuffer,0,320*240),help=0,refresh_screen=1;
    else
      drawhelp(),help=1;

  }
  if (ExKey & GPC_VK_SELECT) {
    if (!help) {
      GpKeyGetEx(&key);while(key!=0)GpKeyGetEx(&key);
      runFileSel();
      gm_memset(vptr[0].ptbuffer, 0, 320*240);
      refresh_screen=1;
    } else { // key define
      int currkey;
      int sil;
      int ExKey2,loop;

      for (loop=0;loop<8;loop++) { // 8 buttons
	currkey=97;
	sil=1;
	while (sil) {
	  gm_memset(vptr[0].ptbuffer,0xff,320*240);
	  if (loop==0) {
	    sprintf(tmpstr,"Select key for left");
	  }
	  if (loop==1) {
	    sprintf(tmpstr,"Select key for right");
	  }
	  if (loop==2) {
	    sprintf(tmpstr,"Select key for up");
	  }
	  if (loop==3) {
	    sprintf(tmpstr,"Select key for down");
	  }
	  if (loop==4) {
	    sprintf(tmpstr,"Select key for A");
	  }
	  if (loop==5) {
	    sprintf(tmpstr,"Select key for B");
	  }
	  if (loop==6) {
	    sprintf(tmpstr,"Select key for left shoulder");
	  }
	  if (loop==7) {
	    sprintf(tmpstr,"Select key for right shoulder");
	  }
	  GpTextOut(NULL, &vptr[0], 10, 10 ,tmpstr , 0x00);
	  sprintf(tmpstr,"< %c >",currkey);
	  GpTextOut(NULL, &vptr[0], 10, 50 ,tmpstr , 0x00);
	  while (GpKeyGetEx(&ExKey2)==0);
	  if (ExKey2 & GPC_VK_LEFT) {
	    currkey--;
	  }
	  if (ExKey2 & GPC_VK_RIGHT) {
	    currkey++;
	  }
	  if (ExKey2 & GPC_VK_FA) {
	    if (loop==0) {
	      but_l=currkey;
	    }
	    if (loop==1) {
	      but_r=currkey;
	    }
	    if (loop==2) {
	      but_u=currkey;
	    }
	    if (loop==3) {
	      but_d=currkey;
	    }
	    if (loop==4) {
	      but_a=currkey;
	    }
	    if (loop==5) {
	      but_b=currkey;
	    }
	    if (loop==6) {
	      but_ls=currkey;
	    }
	    if (loop==7) {
	      but_rs=currkey;
	    }
	    sil=0;
	  }
	  if (currkey<32) { currkey=126; }
	  if (currkey>126) { currkey=32; }
	}
      }

      gm_memset(vptr[0].ptbuffer,0,320*240);
      help=0;
      refresh_screen=1;
    }
  }
}


void screenon()
{
  GpGraphicModeSet(8, NULL);
  GpLcdSurfaceGet( &vptr[0], 0);
  //  GpLcdSurfaceGet( &vptr[1], 1);
  GpSurfaceSet( &vptr[0]);
  GpLcdEnable();

  gm_memset(vptr[0].ptbuffer,0,320*240);
  refresh_screen=1;
}


void screenoff()
{

}


GpMain(void * arg)
{
  GpFatInit();
  GpRelativePathSet("gp:\\gpmm\\gp81\\");

  sprintf(loadgame,"");

  GpClockSpeedChange (33000000,0x24003, 0);

  //struct sigaction sa;
  struct itimerval itv;
  int tmp=1000/50;	/* 50 ints/sec */


 int n;
 n = preInitKbd(9600,33000000);


 screenon();
 GpSetPaletteEntry(0,0,0,0);
 GpSetPaletteEntry(15,255,255,255);

 initmem();

if(scrn_freq<1) scrn_freq=1;
if(scrn_freq>50) scrn_freq=50;

 gp_initRTC();

 mainloop();

 exit(0);
}


loadrom(x)
unsigned char *x;
{
int i;
F_HANDLE in;
 ulong dummy;

 if (GpFileOpen("zx81.rom",OPEN_R,&in)==SM_OK)
  {
  GpFileRead(in,x,8192,&dummy);
  gm_memcpy(x+8192,x,8192);
  GpFileClose(in);
  }
else
  {
  GpError("Couldn't load gpmm\\gp81\\zx81.rom");
  }
}


drawhelp()
{
  gm_memcpy(vptr[0].ptbuffer,zx81kybd,320*240);
}


hacks()
{
/* remove fast mode stuff which screws it all up */
mem[0x4cc]=mem[0x4cd]=mem[0x4ce]=0;

/* replace kybd routine itself, with ld hl,(lastk):ret */
mem[0x2bb]=0x2a; mem[0x2bc]=0x25; mem[0x2bd]=0x40; mem[0x2be]=0xc9;

/* a 'force F/0' at 66h; ok, as we simulate NMI effects by other means */
mem[0x66]=0xcf; mem[0x67]=0x0e;

/* patch save routine */
mem[0x2fc]=0xed; mem[0x2fd]=0xfd;
mem[0x2fe]=0xc3; mem[0x2ff]=0x07; mem[0x300]=0x02;

/* patch load routine */
mem[0x347]=0xeb;
mem[0x348]=0xed; mem[0x349]=0xfc;
mem[0x34a]=0xc3; mem[0x34b]=0x07; mem[0x34c]=0x02;

/* make fast/slow more traceable (directly modify bit 7 rather than
 * bit 6 of CDFLAGS)
 */
mem[0xf29]=0xbe;
mem[0xf2e]=0xfe;

/* do 'out (0),a' when waiting for input (for fast mode display check) */
mem[0x4ca]=0xd3; mem[0x4cb]=0;			/* out (0),a */
mem[0x4cc]=0xcb; mem[0x4cd]=0x46;		/* loop: bit 0,(hl) */
mem[0x4ce]=0x28; mem[0x4cf]=0xfc;		/* jr z,loop */
mem[0x4d0]=0xd3; mem[0x4d1]=1;			/* out (1),a */
mem[0x4d2]=0;	/* nop, as filler */

/* skip some more weird fast mode stuff */
mem[0x2ec]=0xc9;

/* we need the following crock to support 'PAUSE'... */

mem[0xf3b]=0x29;	/* call 0229h */

mem[0x229]=0x22;	/* ld (04034h),hl:nop */
mem[0x22c]=0;

mem[0x23e]=0x2a; mem[0x23f]=0x25; mem[0x240]=0x40;	/* ld hl,(0403bh) */
mem[0x241]=0x23;					/* inc hl */
mem[0x242]=0x3e; mem[0x243]=0x7f;			/* ld a,07fh */
mem[0x244]=0xa4;					/* and h */
mem[0x245]=0xb5;					/* or l */
mem[0x246]=0xc0;					/* ret nz */

mem[0x247]=0x2a;	/* ld hl,(04034h) */
mem[0x248]=0x34;
mem[0x249]=0x40;

mem[0x24a]=0xc3;	/* jp 022dh */
mem[0x24b]=0x2d;
mem[0x24c]=0x02;

/* nop out ld (04034h),hl */
mem[0x23a]=0; mem[0x23a]=0; mem[0x23a]=0;

/* nop out call to 207h */
mem[0xf41]=0; mem[0xf42]=0; mem[0xf43]=0;
}


unsigned int in(h,l)
int h,l;
{
if(l==0xfe)	/* keyboard */
  switch(h)
    {
    case 0xfe:	return(kports[0]);
    case 0xfd:	return(kports[1]);
    case 0xfb:	return(kports[2]);
    case 0xf7:	return(kports[3]);
    case 0xef:	return(kports[4]);
    case 0xdf:	return(kports[5]);
    case 0xbf:	return(kports[6]);
    case 0x7f:	return(kports[7]);
    default:	return(255);
    }

fprintf(stderr," in l=%02X\n",l);
return(255);
}


unsigned int out(h,l,a)
int h,l,a;
{
if(l==0xfd) store(16443,128|fetch(16443));	/* i.e. set slow mode */
if(l==0x00) input_wait=1;	/* waiting for input */
if(l==0x01) input_wait=0;	/* not waiting for input :-) */

return(0);
}


hireschk(ix)
int ix;
{
refresh_screen=1;

if(ix<0x4000)
  hires=0;
else
  {
  int f,g,h,v;
  
  /* now try to find the hi-res screen.
   * we do this by looking for somewhere in ram where there
   * are 192 suitably-spaced EOL chars. The EOL char could theoretically be
   * a few different things, so we just test that it's the same.
   * this is SLOW, but it doesn't matter; this'll only need to
   * happen once, when the hi-res screen is set up.
   *
   * (um, ok, it's a little different now; I just kept hacking until
   * it worked with all three hi-res programs I tried it on :-))
   */
   
  for(f=0x8000-6400;f>=0x4000;f--)
    {
    v=mem[f+32];
    if((v&64)==0) continue;
    hires=1;
    for(g=0;g<192 && hires;g++)
      {
      if(mem[f+33*g]&64) hires=0;
      if(mem[f+32+33*g]!=v) hires=0;
      }
    if(hires)
      {
      hiscrn=f;
      return(0);
      }
    }
  }
}


hireschki(i)
int i;
{
iptr=mem+(i<<8);
}


/* draw "hi-res" screen */
drawhires()
{
int x,y,c,d,inv,mask,a,ofs;
unsigned char *tmp,*ptr;

ptr=mem+hiscrn;

ofs=0;
for(y=0;y<192;y++,ptr++)
  {
  for(x=0;x<32;x++,ptr++,ofs++)
    {
    c=*ptr;
    
    if((himap[ofs]=c)!=himap_old[ofs] || refresh_screen)
      {
      inv=(c&128); c&=63;
      
      tmp=vptr[0].ptbuffer+(32*240+216)-y+(x*8*240);
      d=iptr[c*8]; if(inv) d^=255;
      mask=128;
      for(a=0;a<8;a++,mask>>=1) {
        *tmp=(d&mask)?0:15;
	tmp+=240;
      }
      }
    }
  }

/* now, copy new to old for next time */
gm_memcpy(himap_old,himap,32*192);
refresh_screen=0;
}


/* redraw the screen */
update_scrn()
{
static int count=0;
int x,y,a,b,c,d,inv,x2,mask;
unsigned char *ptr,*cptr,*tmp;
int fast=(!(fetch(16443)&128) && !input_wait);
int pasteol;
int ofs;

/* only do it every 1/25th */
count++;
if(count<(lowref?25:scrn_freq)) return(0); else count=0;

/* only if not showing help screen */
if(help) return(0);

if(hires && !fast)
  {
    drawhires();
    return(0);
  }

ptr=mem+fetch2(16396);	/* D_FILE */
/* don't bother if it's junk */
if(ptr-mem<0 || ptr-mem>0xf000) return(0);
ptr++;	/* skip first HALT */

cptr=mem+0x1e00;

ofs=0;
for(y=0;y<24;y++,ptr++)
  {
  pasteol=0;
  for(x=0;x<32;x++,ptr++,ofs++)
    {
    if(pasteol || (*ptr)==0x76)
      {
      pasteol=1;
      c=0;	/* space */
      ptr--;
      }
    else
      c=*ptr;
    
    if((chrmap[ofs]=c)!=chrmap_old[ofs] || refresh_screen)
      {
      inv=(c&128); c&=63;
      
      for(b=0;b<8;b++)
        {
	  // tmp=vptr[0].ptbuffer+4*320+32+(y*8+b)*320+x*8;
	  tmp=vptr[0].ptbuffer+(32*240+216)-(y*8+b)+(x*8*240);
        d=cptr[c*8+b]; if(inv) d^=255;
        mask=128;
        for(a=0;a<8;a++,mask>>=1) {
          *tmp=(d&mask)?0:15;
	  tmp+=240;
	}
        }
      }
    }
  }

/* now, copy new to old for next time */
gm_memcpy(chrmap_old,chrmap,768);
refresh_screen=0;

}


update_kybd()
{
int y,b;
int lastk0=0,lastk1=0;	/* values for lastk and lastk+1 */
static int oldlk0=0,oldlk1=0;
int f;
unsigned short tmp;

/* XXX want to support DEL etc. */

/* bump frames */
f=(fetch2(16436)&32767);
if(f>0) f--;
store2(16436,(fetch2(16436)&32768)|f);

/* set max T-states per 1/50th to get the right speed */

if(topspeed)
  tsmax=(1<<30);
else
  if(fetch(16443)&128)
    {
    /* slow mode - 0.8 MHz, 16000 T-states per 1/50th */
      tsmax=16000;
    }
  else
    {
    /* fast mode - 3.2 MHz, 64000 T-states per 1/50th */
          tsmax=64000;
    }

for(y=0;y<8;y++)		/* 8 half-rows */
  {
    b=kports[y];

  /* contribute to lastk if key was pressed */
  if((b&31)!=31)
    {
    /* set y bit in lastk0 if not shift */
    if(!(y==0 && ((b&31)|1)==31)) lastk0|=(1<<y);
    
    /* now set pos-in-half-row bit(s) in lastk1 */
    b=(((~b)&31)<<1);
    /* move bit 1 of b back down to bit 0 if it's shift bit */
    if(y==0) b=((b&0xfc)|((b&2)>>1));
  
    lastk1|=b;
    }
  }

/* see if this is different to prev. lastk[01] */
store(16443,(fetch(16443)&0xfe)|
	((((lastk1&0xfe)|lastk0) && (oldlk0!=lastk0 || oldlk1!=lastk1))?1:0));


oldlk0=lastk0; oldlk1=lastk1;

/* fix lastk values */
lastk0^=255; lastk1^=255;

/* put lastk stuff in place */
store(16421,lastk0);
store(16422,lastk1);

/* and FWIW, set MARGIN */
store(16424,55);
}


/* the ZX81 char is used to index into this, to give the ascii.
 * awkward chars are mapped to '_' (underscore), and the ZX81's
 * all-caps is converted to lowercase.
 * WARNING: this only covers 0<=char<=63!
 */
char zx2ascii[64]={
/* 000-009 */ ' ', '_', '_', '_', '_', '_', '_', '_', '_', '_', 
/* 010-019 */ '_', '\'','#', '$', ':', '?', '(', ')', '>', '<', 
/* 020-029 */ '=', '+', '-', '*', '/', ';', ',', '.', '0', '1', 
/* 030-039 */ '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 
/* 040-049 */ 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 
/* 050-059 */ 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 
/* 060-063 */ 'w', 'x', 'y', 'z'
};

save_p(a)
int a;
{
static char fname[128];
unsigned char *ptr=mem+a,*dptr=fname;
FILE *out;

/* so the filename is at ptr, in the ZX81 char set, with the last char
 * having bit 7 set. First, get that name in ASCII.
 */

gm_memset(fname,0,sizeof(fname));
do
  *dptr++=zx2ascii[(*ptr)&63];
while((*ptr++)<127);

/* add '.p' */
strcat(fname,".p");

/* now open the file */
if((out=fopen(fname,"wb"))==NULL)
  {
  forceok=1;
  return(0);
  }

/* work out how much to write, and write it.
 * we need to write from 0x4009 to E_LINE inclusive.
 */
fwrite(mem+0x4009,1,fetch2(16404)-0x4009+1,out);

fclose(out);
}

load_p(a)
int a;
{
  ulong dummy;
static char fname[15];
unsigned char *ptr=mem+(a&32767),*dptr=fname;
F_HANDLE in;

gm_memset(fname,0,sizeof(fname));

if(a&32768) 	/* they did LOAD "" */
  {
   sprintf(fname,"%s",loadgame);
   sprintf(loadgame,"");
  } else {

/* so the filename is at ptr, in the ZX81 char set, with the last char
 * having bit 7 set. First, get that name in ASCII.
 */

/* test for Xtender-style LOAD " STOP " to quit */
if(*ptr==227) dontpanic();

do
  *dptr++=zx2ascii[(*ptr)&63];
while((*ptr++)<127);

  }
/* add '.ppp' (not .p 'cos we really should stick to the 8.3 standard) */
 strcat(fname,".ppp");


/* now open the file */
 if (GpFileOpen(fname,OPEN_R,&in)!=SM_OK)
  {
  forceok=1;
  return(0);
  }

/* just read it all */
GpFileRead(in,mem+0x4009,16384,&dummy);

GpFileClose(in);

}


fix_tstates()
{
tstates=0;
 while(interrupted==0);
}


do_interrupt()
{
  int ExKey;
  if(booting){
    if(input_wait) {
      booting=0;
    } else {
      return(0);
    }
  }
  update_scrn();
  update_gpkbd();
  GpKeyGetEx(&ExKey);
  if (ExKey) {
    update_gppress(ExKey);
  }
  update_kybd();

  if (keyhold>0) {
    keyhold--; 
  } else {
    keyrelease();
  }
if(interrupted<2) interrupted=0;
}
